from .requester import Requester
from .logger import Logger
from .exceptions import (
    OSApiAuthenticationFail,
    OSSessionRefreshFail,
    OSFailedToLocateSession,
)

from .endpoints import *  # noqa: F403

import requests


class API:
    def __init__(
        self, base_url: str, username: str, password: str, rate_limit: bool = True
    ) -> None:
        """ Connection to Odin API, all interactions with the api are here.

        Args:
            base_url (str): Base url of your odin instance api.
            username (str): Username used when logging into odin account.
            password (str): Password used when logging into odin account stored as virtual environment.
            rate_limit (bool): Enables (True) or Disables (False) rate limiting to 5 calls per second. Defaults to True.
            
        Vars: 
            authorised (bool): Boolean value to indicate if api is authorised.\
            token (str): Token string returned from odin api.
        """

        self.base_url = base_url
        self.username = username
        self._password = password
        self.rate_limit = rate_limit

        self.authorised = False
        self.token = ""

        self.logger = Logger.get_instance(self.username)
        self._requester = Requester(self.base_url, self.rate_limit, self.logger)

        self.administrators = Administrators()
        self.alternate_numbers = AlternateNumbers()
        self.authentication = Authentication()
        self.auto_attendants = AutoAttendants()
        self.call_centers = CallCenters()
        self.call_forwarding_always = CallForwardingAlways()
        self.call_forwarding_busy = CallForwardingBusy()
        self.call_forwarding_no_answer = CallForwardingNoAnswer()
        self.call_forwarding_not_reachable = CallForwardingNotReachable()
        self.call_forwarding_selective = CallForwardingSelective()
        self.call_pickup = CallPickup()
        self.call_processing_policies = CallProcessingPolicies()
        self.call_records = CallRecords()
        self.devices = Devices()
        self.dns = DNs()
        self.groups = Groups()
        self.emergency_zones = EmergencyZones()
        self.do_not_disturb = DoNotDisturb()
        self.hunt_groups = HuntGroups()
        self.service_providers = ServiceProviders()
        self.services = Services()
        self.shared_call_appearance = SharedCallAppearance()
        self.schedules = Schedules()
        self.reports = Reports()
        self.regsitration = Registration()
        self.password_generate = PasswordGenerate()
        self.trunk_groups = TrunkGroups()
        self.users = Users()

    def authenticate(self) -> bool:
        """Authenticates session with username and password supplied by user.

        Raises:
            OSApiAuthenticationFail: Raised if authenticaion fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """

        endpoint = "/auth/token"

        payload = {"username": self.username, "password": self._password}

        try:
            response = self._requester.post(endpoint, data=payload)
            self._update_requester(response)
            return True
        except requests.exceptions.HTTPError:
            raise OSApiAuthenticationFail()

    def refresh_authorisation(self) -> bool:
        """Re-authenticates the session with the API. Used if API key is to expire.

        Raises:
            OSSessionRefreshFail: Raised if authentication fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """

        endpoint = "/auth/session"

        try:
            response = self._requester.put(endpoint)
            self._update_requester(response)
            return True
        except requests.exceptions.HTTPError:
            raise OSSessionRefreshFail()

    def get_auth_details(self):
        """Gets current session details.

        Raises:
            OSFailedToLocateSession: Raised when session details can't be found.
                Most likely because session has expired.

        Returns:
            Dict: Current session details.
        """

        endpoint = "/auth/session"

        try:
            return self._requester.get(endpoint)
        except requests.exceptions.HTTPError:
            raise OSFailedToLocateSession()

    def update_api(
        self,
        base_url: str = None,
        username: str = None,
        password: str = None,
        rate_limit: bool = None,
    ):
        """Updates the API with new details.

        Args:
            base_url (str, optional): Base url of your odin instance api. Defaults to None.
            username (str, optional): Username used when logging into odin account. Defaults to None.
            password (str, optional): Password used when logging into odin account stored as virtual environment. Defaults to None.
            rate_limit (bool, optional): Enables (True) or Disables (False) rate limiting to 5 calls per second. Defaults to None.
        """

        if base_url:
            self.base_url = base_url
            self._requester.base_url = base_url
        if username:
            self.username = username
            self.logger = Logger.get_instance(self.username)
            self._requester.logger = self.logger
        if password:
            self._password = password
        if rate_limit:
            self.rate_limit = rate_limit
            self._requester.rate_limit = rate_limit

    def _update_requester(self, session_response: requests.models.Response):
        """When authenticating or re-auth update requester with token so it can make
        api calls

        Args:
            session_response (request): Requests mod response.
        """

        self.token = session_response["token"]
        self._requester.headers["Authorization"] = f"Bearer {self.token}"
        self.authorised = True

    def __str__(self) -> str:
        return (
            f"API - url: {self.base_url}, username: {self.username} "
            f"Authenticated: {self.authorised}"
        )



from ..exceptions import OSExtensionNotFound, OSRangeFault


def retrieve_extensions(api, service_provider_id: str, group_id: str) -> list:
    extensions = []

    dataset = (
        api.users.get_users(service_provider_id, group_id)
        + api.hunt_groups.get_group_hunt_groups(service_provider_id, group_id)
        + api.call_centers.get_group_call_centers(service_provider_id, group_id)
        + api.auto_ettendants.get_auto_attendants(service_provider_id, group_id)
    )

    for data in dataset:
        if not data["extension"]:
            continue

        extensions.append(int(data["extension"]))

    return extensions if extensions else None


def main(
    api, service_provider_id: str, group_id: str, range_start: int, range_end: int
):
    """Retrieves The Lowest Free Extension Available In The Designated Group Passed."""

    # Raise Exception If Range Start Is Greater Than Range End
    if range_start > range_end:
        raise OSRangeFault

    # Retrieve List Of Occupied Extensions Within The Group
    extensions = retrieve_extensions(
        api,
        service_provider_id,
        group_id,
    )

    for extension in range(range_start, range_end + 1):
        if extension not in extensions:
            return {"extension": extension}

    raise OSExtensionNotFound



from .base_endpoint import BaseEndpoint


class Administrators(BaseEndpoint):
    def __init__(self):
        super().__init__()

    # GET

    # POST
    def post_group_admin(
        self,
        service_provider_id: str,
        group_id: str,
        user_id: str,
        password: str,
        payload: dict = {},
    ):
        """Builds a group-level administrator.

        Args:
            service_provider_id (str): Service provider ID where the admin should be built.
            group_id (str): Group ID where the admin should be built
            user_id (str): User ID of the admin.
            password (str): Password for the administrator profile. Note get.password_generate() can be used to get this.
            payload (dict, optional): Admin configuration data.

        Returns:
            Dict: Returns the admin profile.
        """

        endpoint = "/groups/admins"

        payload["serviceProviderId"] = service_provider_id
        payload["groupId"] = group_id
        payload["userId"] = user_id
        payload["password"] = password

        return self._requester.post(endpoint, data=payload)

    def post_group_admin_policies_bulk(self, user_ids: list, policy_config: dict):
        """Applies policy settings to multiple group administrators.

        Note: See docs for formatting of parameters.

        Args:
            user_ids (list): User IDs of admins to apply policy to.
            policy_config (dict): Policy settings to apply to target users.

        Returns:
            Dict: Returns admins and policy applied.
        """
        endpoint = "/groups/admins/policies/bulk"

        data = {"users": [{"userId": user} for user in user_ids], "data": policy_config}

        self._requester.post(endpoint, data=data)

    # PUT
    # DELETE
    


import requests
import json
from ratelimit import limits, sleep_and_retry
from .exceptions import OSApiResponseError


class Requester:
    __instance = None  # Class variable to hold the singleton instance

    @staticmethod
    def get_instance(base_url=None, rate_limit=None, logger=None):
        if Requester.__instance is None:
            Requester(base_url, rate_limit, logger)
        return Requester.__instance

    def __init__(self, base_url, rate_limit, logger):
        """
        Initialize the Requester with default values.

        NOTE: This object is a singleton and can't be instatiated more than once.
        """

        if Requester.__instance is not None:
            raise Exception("Singleton cannot be instantiated more than once!")
        else:
            self.base_url = base_url
            self.rate_limit = rate_limit
            self.logger = logger
            self.headers = {
                "Authorization": "",
                "Content-Type": "application/json",
            }

            Requester.__instance = self

    # get, post, put, delete methods takes in data and params and returns method type to _request
    def get(self, endpoint, data=None, params=None):
        return self._request(requests.get, endpoint, data, params)

    def post(self, endpoint, data=None):
        return self._request(requests.post, endpoint, data)

    def put(self, endpoint, data=None):
        return self._request(requests.put, endpoint, data)

    def delete(self, endpoint, data=None, params=None):
        return self._request(requests.delete, endpoint, data, params)

    def _request(self, method, endpoint, data=None, params=None):
        """Handles an API request with or without rate limiting.

        Args:
            method (request obj): Request module method type either GET, POST, PUT, DELETE
            endpoint (str): Specific API endpoint for functionality
            data (dict, optional): Python dict used in payload data if needed. Defaults to None.
            params (dict, optional): Parameters used in endpoint if needed. Defaults to None.

        Returns:
            API data: Data returned from API on completion of API call.
        """

        # if rate limiting is enabled uses _rate_limited_request where limiting is in place.
        if self.rate_limit:
            return self._rate_limited_request(method, endpoint, data, params)
        else:
            response = method(
                url=self.base_url + endpoint,
                headers=self.headers,
                data=json.dumps(data if data is not None else {}),
                params=(params if params is not None else {}),
            )

            # if logger used log request
            if self.logger:
                self.logger._log_request(
                    endpoint=endpoint, response_code=response.status_code
                )

            # flags errors if any returned from the API
            try:
                response.raise_for_status()
            except requests.exceptions.RequestException:
                raise OSApiResponseError(response)
            else:
                return response.json()

    @sleep_and_retry
    @limits(calls=5, period=1)
    def _rate_limited_request(self, method, endpoint, data=None, params=None):
        """Handles an API request with rate limiting."""

        response = method(
            url=self.base_url + endpoint,
            headers=self.headers,
            data=json.dumps(data if data is not None else {}),
            params=(params if params is not None else {}),
        )

        # if logger used log request
        if self.logger:
            self.logger._log_request(
                endpoint=endpoint, response_code=response.status_code
            )

        # flags errors if any returned from the API
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise OSApiResponseError(response)
        else:
            return response.json()
